#pragma once

#include <cppcore/misc/string_builder.h>

#include "registry.h"
#include "misc/exception.h"
#include "traits/default_type.h"

namespace cpprtti
{

    /* registry */

    template<typename T_type>
    auto& registry::get()
        { return get<T_type, default_type_t<T_type>>(); }

    template<typename T_type>
    const auto& registry::get() const
    {
        using impl_type = default_type_t<T_type>;

        auto * t = find<T_type>();

        if (!t)
        {
            throw exception(
                error_code::registry_type_not_found,
                cppcore::string_builder()
                    << "The requested type could not be found: "
                    << get_type_name<T_type>());
        }

        auto * ret = dynamic_cast<const impl_type *>(t);
        if (!ret)
        {
            throw exception(
                error_code::registry_type_does_not_match,
                cppcore::string_builder()
                    << "The type does not match the expected implementation: "
                    << get_type_name<impl_type>());
        }

        return *ret;
    }

    template<typename T_type, typename T_impl>
    auto& registry::get()
    {
        auto * t = find<T_type>();
        if (!t)
            t = &create<T_impl>();

        auto * ret = dynamic_cast<T_impl *>(t);
        if (!ret)
        {
            throw exception(
                error_code::registry_type_does_not_match,
                cppcore::string_builder()
                    << "The type does not match the expected implementation: "
                    << get_type_name<T_impl>());
        }

        return * ret;
    }

    template<typename T_type>
    type * registry::find()
        { return find(get_type_id<cppmp::decay_t<T_type>>()); }

    template<typename T_type>
    const type * registry::find() const
        { return find(get_type_id<cppmp::decay_t<T_type>>()); }

    type * registry::find(type_id id)
    {
        auto it = _id_map.find(id);
        return it == _id_map.end()
            ? nullptr
            : it->second.get();
    }

    const type * registry::find(type_id id) const
    {
        auto it = _id_map.find(id);
        return it == _id_map.end()
            ? nullptr
            : it->second.get();
    }

    type * registry::find(const std::string& name)
    {
        auto it = _name_map.find(name);
        return it == _name_map.end()
            ? nullptr
            : it->second.get();
    }

    const type * registry::find(const std::string& name) const
    {
        auto it = _name_map.find(name);
        return it == _name_map.end()
            ? nullptr
            : it->second.get();
    }

    template<typename T_impl>
    T_impl& registry::create()
    {
        auto impl = std::make_shared<T_impl>(*this);

        auto id_it = _id_map.find(impl->id());
        if (id_it != _id_map.end())
        {
            throw exception(
                error_code::registry_id_already_exsists,
                cppcore::string_builder()
                    << "A type with the id '"
                    << impl->id()
                    << "' is already registered!");
        }

        auto name_it = _name_map.find(impl->name());
        if (name_it != _name_map.end())
        {
            throw exception(
                error_code::registry_name_already_exsists,
                cppcore::string_builder()
                    << "A type with the name '"
                    << impl->name()
                    << "' is already registered!");
        }

        _id_map.emplace_hint(id_it, impl->id(), impl);
        _name_map.emplace_hint(name_it, impl->name(), impl);

        return *impl;
    }

}
